/**
 * Performs single click mouse gesture.
 *
 * @this {WindowsDriver}
 * @param {string} [elementId] Hexadecimal identifier of the element to click on.
 * If this parameter is missing then given coordinates will be parsed as absolute ones.
 * Otherwise they are parsed as relative to the top left corner of this element.
 * @param {number} [x] Integer horizontal coordinate of the click point. Both x and y coordinates
 * must be provided or none of them if elementId is present. In such case the gesture
 * will be performed at the center point of the given element.
 * @param {number} [y] Integer vertical coordinate of the click point. Both x and y coordinates
 * must be provided or none of them if elementId is present. In such case the gesture
 * will be performed at the center point of the given element.
 * @param {'left' | 'middle' | 'right' | 'back' | 'forward'} [button=left] Name of
 * the mouse button to be clicked. An exception is thrown if an unknown button name
 * is provided.
 * @param {string[]|string} [modifierKeys] List of possible keys or a single key name to
 * depress while the click is being performed. Supported key names are: Shift, Ctrl, Alt, Win.
 * For example, in order to keep Ctrl+Alt depressed while clicking, provide the value of
 * ['ctrl', 'alt']
 * @param {number} [durationMs] The number of milliseconds to wait between pressing
 * and releasing the mouse button. By default no delay is applied, which simulates a
 * regular click.
 * @param {number} [times=1] How many times the click must be performed.
 * @param {number} [interClickDelayMs=100] Duration od the pause between each
 * click gesture. Only makes sense if `times` is greater than one.
 * @throws {Error} If given options are not acceptable or the gesture has failed.
 */
export function windowsClick(this: import("../driver").WindowsDriver, elementId?: string, x?: number, y?: number, button?: "left" | "middle" | "right" | "back" | "forward", modifierKeys?: string[] | string, durationMs?: number, times?: number, interClickDelayMs?: number): Promise<void>;
/**
 * Performs horizontal or vertical scrolling with mouse wheel.
 *
 * @this {WindowsDriver}
 * @param {string} [elementId] Hexadecimal identifier of the element to scroll.
 * If this parameter is missing then given coordinates will be parsed as absolute ones.
 * Otherwise they are parsed as relative to the top left corner of this element.
 * @param {number} [x] Integer horizontal coordinate of the scroll point. Both x and y coordinates
 * must be provided or none of them if elementId is present. In such case the gesture
 * will be performed at the center point of the given element.
 * @param {number} [y] Integer vertical coordinate of the scroll point. Both x and y coordinates
 * must be provided or none of them if elementId is present. In such case the gesture
 * will be performed at the center point of the given element.
 * @param {number} [deltaX] Integer horizontal scroll delta. Either this value
 * or deltaY must be provided, but not both.
 * @param {number} [deltaY] Integer vertical scroll delta. Either this value
 * or deltaX must be provided, but not both.
 * @param {string[]|string} [modifierKeys] List of possible keys or a single key name to
 * depress while the scroll is being performed. Supported key names are: Shift, Ctrl, Alt, Win.
 * For example, in order to keep Ctrl+Alt depressed while clicking, provide the value of
 * ['ctrl', 'alt']
 * @throws {Error} If given options are not acceptable or the gesture has failed.
 */
export function windowsScroll(this: import("../driver").WindowsDriver, elementId?: string, x?: number, y?: number, deltaX?: number, deltaY?: number, modifierKeys?: string[] | string): Promise<void>;
/**
 * Performs drag and drop mouse gesture.
 *
 * @this {WindowsDriver}
 * @param {string} [startElementId] Hexadecimal identifier of the element to start the drag from.
 * If this parameter is missing then given coordinates will be parsed as absolute ones.
 * Otherwise they are parsed as relative to the top left corner of this element.
 * @param {number} [startX] Integer horizontal coordinate of the drag start point. Both startX
 * and startY coordinates must be provided or none of them if elementId is present. In such case the gesture
 * will be performed at the center point of the given element.
 * @param {number} [startY] Integer vertical coordinate of the drag start point. Both startX and
 * startY coordinates must be provided or none of them if elementId is present. In such case the gesture
 * will be performed at the center point of the given element.
 * @param {string} [endElementId] Hexadecimal identifier of the element to end the drag on.
 * If this parameter is missing then given coordinates will be parsed as absolute ones.
 * Otherwise they are parsed as relative to the top left corner of this element.
 * @param {number} [endX] Integer horizontal coordinate of the drag end point. Both endX and endY coordinates
 * must be provided or none of them if elementId is present. In such case the gesture
 * will be performed at the center point of the given element.
 * @param {number} [endY] Integer vertical coordinate of the drag end point. Both endX and endY coordinates
 * must be provided or none of them if elementId is present. In such case the gesture
 * will be performed at the center point of the given element.
 * @param {string[]|string} [modifierKeys] List of possible keys or a single key name to
 * depress while the drag is being performed. Supported key names are: Shift, Ctrl, Alt, Win.
 * For example, in order to keep Ctrl+Alt depressed while clicking, provide the value of
 * ['ctrl', 'alt']
 * @param {number} [durationMs=5000] The number of milliseconds to wait between pressing
 * the left mouse button and moving the cursor to the ending drag point.
 * @throws {Error} If given options are not acceptable or the gesture has failed.
 */
export function windowsClickAndDrag(this: import("../driver").WindowsDriver, startElementId?: string, startX?: number, startY?: number, endElementId?: string, endX?: number, endY?: number, modifierKeys?: string[] | string, durationMs?: number): Promise<void>;
/**
 * Performs hover mouse gesture.
 *
 * @this {WindowsDriver}
 * @param {string} [startElementId] Hexadecimal identifier of the element to start the hover from.
 * If this parameter is missing then given coordinates will be parsed as absolute ones.
 * Otherwise they are parsed as relative to the top left corner of this element.
 * @param {number} [startX] Integer horizontal coordinate of the hover start point. Both startX
 * and startY coordinates must be provided or none of them if elementId is present. In such case the gesture
 * will be performed at the center point of the given element.
 * @param {number} [startY] Integer vertical coordinate of the hover start point. Both startX and
 * startY coordinates must be provided or none of them if elementId is present. In such case the gesture
 * will be performed at the center point of the given element.
 * @param {string} [endElementId] Hexadecimal identifier of the element to end the hover on.
 * If this parameter is missing then given coordinates will be parsed as absolute ones.
 * Otherwise they are parsed as relative to the top left corner of this element.
 * @param {number} [endX] Integer horizontal coordinate of the hover end point. Both endX and endY coordinates
 * must be provided or none of them if elementId is present. In such case the gesture
 * will be performed at the center point of the given element.
 * @param {number} [endY] Integer vertical coordinate of the hover end point. Both endX and endY coordinates
 * must be provided or none of them if elementId is present. In such case the gesture
 * will be performed at the center point of the given element.
 * @param {string[]|string} [modifierKeys] List of possible keys or a single key name to
 * depress while the hover is being performed. Supported key names are: Shift, Ctrl, Alt, Win.
 * For example, in order to keep Ctrl+Alt depressed while hovering, provide the value of
 * ['ctrl', 'alt']
 * @param {number} [durationMs=500] The number of milliseconds between
 * moving the cursor from the starting to the ending hover point.
 * @throws {Error} If given options are not acceptable or the gesture has failed.
 */
export function windowsHover(this: import("../driver").WindowsDriver, startElementId?: string, startX?: number, startY?: number, endElementId?: string, endX?: number, endY?: number, modifierKeys?: string[] | string, durationMs?: number): Promise<void>;
/**
 * @typedef {Object} KeyAction
 * @property {number} pause Allows to set a delay in milliseconds between key input series.
 * Either this property or `text` or `virtualKeyCode` must be provided.
 * @property {string} text Non-empty string of Unicode text to type.
 * Either this property or `pause` or `virtualKeyCode` must be provided.
 * @property {number} virtualKeyCode Valid virtual key code. The list of supported key codes
 * is available at https://learn.microsoft.com/en-us/windows/win32/inputdev/virtual-key-codes
 * Either this property or `pause` or `text` must be provided.
 * @property {boolean?} down [undefined] This property only makes sense in combination with `virtualKeyCode`.
 * If set to `true` then the corresponding key will be depressed, `false` - released. By default
 * the key is just pressed once.
 * ! Do not forget to release depressed keys in your automated tests.
 */
/**
 * Performs customized keyboard input.
 *
 * @this {WindowsDriver}
 * @param {KeyAction[] | KeyAction} actions One or more key actions.
 * @throws {Error} If given options are not acceptable or the gesture has failed.
 */
export function windowsKeys(this: import("../driver").WindowsDriver, actions: KeyAction[] | KeyAction): Promise<void>;
export type KeyAction = {
    /**
     * Allows to set a delay in milliseconds between key input series.
     * Either this property or `text` or `virtualKeyCode` must be provided.
     */
    pause: number;
    /**
     * Non-empty string of Unicode text to type.
     * Either this property or `pause` or `virtualKeyCode` must be provided.
     */
    text: string;
    /**
     * Valid virtual key code. The list of supported key codes
     * is available at https://learn.microsoft.com/en-us/windows/win32/inputdev/virtual-key-codes
     * Either this property or `pause` or `text` must be provided.
     */
    virtualKeyCode: number;
    /**
     * [undefined] This property only makes sense in combination with `virtualKeyCode`.
     * If set to `true` then the corresponding key will be depressed, `false` - released. By default
     * the key is just pressed once.
     * ! Do not forget to release depressed keys in your automated tests.
     */
    down: boolean | null;
};
export type WindowsDriver = import("../driver").WindowsDriver;
//# sourceMappingURL=gestures.d.ts.map