"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getWADExecutablePath = void 0;
exports.isAdmin = isAdmin;
const lodash_1 = __importDefault(require("lodash"));
const support_1 = require("appium/support");
const path_1 = __importDefault(require("path"));
const teen_process_1 = require("teen_process");
const logger_1 = require("./logger");
const registry_1 = require("./registry");
const utils_1 = require("./utils");
const POSSIBLE_WAD_INSTALL_ROOTS = [
    process.env['ProgramFiles(x86)'],
    process.env.ProgramFiles,
    `${process.env.SystemDrive || 'C:'}\\\\Program Files`,
];
const WAD_EXE_NAME = 'WinAppDriver.exe';
const UNINSTALL_REG_ROOT = 'HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall';
const REG_ENTRY_VALUE = 'Windows Application Driver';
const REG_ENTRY_KEY = 'DisplayName';
const REG_ENTRY_TYPE = 'REG_SZ';
const INST_LOCATION_SCRIPT_BY_GUID = (guid) => `
Set installer = CreateObject("WindowsInstaller.Installer")
Set session = installer.OpenProduct("${guid}")
session.DoAction("CostInitialize")
session.DoAction("CostFinalize")
WScript.Echo session.Property("INSTALLFOLDER")
`.replace(/\n/g, '\r\n');
async function fetchMsiInstallLocation(installerGuid) {
    const tmpRoot = await support_1.tempDir.openDir();
    const scriptPath = path_1.default.join(tmpRoot, 'get_wad_inst_location.vbs');
    try {
        await support_1.fs.writeFile(scriptPath, INST_LOCATION_SCRIPT_BY_GUID(installerGuid), 'latin1');
        const { stdout } = await (0, utils_1.shellExec)('cscript.exe', ['/Nologo', scriptPath]);
        return lodash_1.default.trim(stdout);
    }
    finally {
        await support_1.fs.rimraf(tmpRoot);
    }
}
class WADNotFoundError extends Error {
}
exports.getWADExecutablePath = lodash_1.default.memoize(async function getWADInstallPath() {
    const wadPath = process.env.APPIUM_WAD_PATH ?? '';
    if (await support_1.fs.exists(wadPath)) {
        logger_1.log.debug(`Loaded WinAppDriver path from the APPIUM_WAD_PATH environment variable: ${wadPath}`);
        return wadPath;
    }
    // TODO: WAD installer should write the full path to it into the system registry
    const pathCandidates = POSSIBLE_WAD_INSTALL_ROOTS
        // remove unset env variables
        .filter(Boolean)
        // construct full path
        // @ts-ignore The above filter does the job
        .map((root) => path_1.default.resolve(root, REG_ENTRY_VALUE, WAD_EXE_NAME));
    for (const result of pathCandidates) {
        if (await support_1.fs.exists(result)) {
            return result;
        }
    }
    logger_1.log.debug('Did not detect WAD executable at any of the default install locations');
    logger_1.log.debug('Checking the system registry for the corresponding MSI entry');
    try {
        const uninstallEntries = await (0, registry_1.queryRegistry)(UNINSTALL_REG_ROOT);
        const wadEntry = uninstallEntries.find(({ key, type, value }) => key === REG_ENTRY_KEY && value === REG_ENTRY_VALUE && type === REG_ENTRY_TYPE);
        if (wadEntry) {
            logger_1.log.debug(`Found MSI entry: ${JSON.stringify(wadEntry)}`);
            const installerGuid = lodash_1.default.last(wadEntry.root.split('\\'));
            // WAD MSI installer leaves InstallLocation registry value empty,
            // so we need to be hacky here
            const result = path_1.default.join(await fetchMsiInstallLocation(installerGuid), WAD_EXE_NAME);
            logger_1.log.debug(`Checking if WAD exists at '${result}'`);
            if (await support_1.fs.exists(result)) {
                return result;
            }
            logger_1.log.debug(result);
        }
        else {
            logger_1.log.debug('No WAD MSI entries have been found');
        }
    }
    catch (e) {
        if (e.stderr) {
            logger_1.log.debug(e.stderr);
        }
        logger_1.log.debug(e.stack);
    }
    throw new WADNotFoundError(`${WAD_EXE_NAME} has not been found in any of these ` +
        `locations: ${pathCandidates}. Use the following driver script to install it: ` +
        `'appium driver run windows install-wad <optional_wad_version>'. ` +
        `Check https://github.com/microsoft/WinAppDriver/releases to list ` +
        `available server versions or drop the '<optional_wad_version>' argument to ` +
        `install the latest stable one.`);
});
async function isAdmin() {
    try {
        await (0, teen_process_1.exec)('fsutil.exe', ['dirty', 'query', process.env.SystemDrive || 'C:']);
        return true;
    }
    catch {
        return false;
    }
}
;
//# sourceMappingURL=installer.js.map