"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.queryRegistry = queryRegistry;
exports.parseRegQueryOutput = parseRegQueryOutput;
const lodash_1 = __importDefault(require("lodash"));
const utils_1 = require("./utils");
const REG = 'reg.exe';
const ENTRY_PATTERN = /^\s+(\w+)\s+([A-Z_]+)\s*(.*)/;
function parseRegEntries(root, block) {
    return (lodash_1.default.isEmpty(block) || lodash_1.default.isEmpty(root))
        ? []
        : block.reduce((acc, line) => {
            const match = ENTRY_PATTERN.exec(line);
            if (match) {
                acc.push({ root, key: match[1], type: match[2], value: match[3] || '' });
            }
            return acc;
        }, []);
}
function parseRegQueryOutput(output) {
    const result = [];
    let root;
    let regEntriesBlock = [];
    for (const line of output.split('\n').map(lodash_1.default.trimEnd)) {
        if (!line) {
            continue;
        }
        const curIndent = line.length - lodash_1.default.trimStart(line).length;
        if (curIndent === 0) {
            result.push(...parseRegEntries(root, regEntriesBlock));
            root = line;
            regEntriesBlock = [];
        }
        else {
            regEntriesBlock.push(line);
        }
    }
    result.push(...parseRegEntries(root, regEntriesBlock));
    return result;
}
/**
 * @typedef {Object} RegEntry
 * @property {string} root Full path to the registry branch, for example
 * HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\DirectDrawEx
 * @property {string} key The registry key name
 * @property {string} type One of possible registry value types, for example
 * REG_DWORD or REG_SZ
 * @property {string} value The actual value. Could be empty
 */
/**
 * Lists registry tree (e.g. recursively) under the given root node.
 * The lookup is done under the same registry branch that the current process
 * system architecture.
 *
 * @param {string} root The registry key name, which consists of two parts:
 * - The root key: HKLM | HKCU | HKCR | HKU | HKCC
 * - The subkey under the selected root key, for example \Software\Microsoft
 * @returns {Promise<RegEntry[]>} List of matched RegEntry instances or an empty list
 * if either no entries were found under the given root or the root does not exist.
 */
async function queryRegistry(root) {
    let stdout;
    try {
        ({ stdout } = await (0, utils_1.shellExec)(REG, ['query', root, '/s']));
    }
    catch {
        return [];
    }
    return parseRegQueryOutput(stdout);
}
//# sourceMappingURL=registry.js.map